NETBOX_DEVICES_API_CALL = "/api/dcim/devices/?format=json"
NETBOX_DEVICES_API_RESPONSE = {
    "count": 37,
    "next": None,
    "previous": None,
    "results": [
        {
            "id": 1,
            "url": "http://localhost:55891/api/dcim/devices/1/?format=json",
            "display": "dmi01-akron-rtr01",
            "name": "dmi01-akron-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/sites/2/?format=json",
                "display": "DM-Akron",
                "name": "DM-Akron",
                "slug": "dm-akron",
            },
            "location": None,
            "rack": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/racks/1/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:26:44.813000Z",
        },
        {
            "id": 14,
            "url": "http://localhost:55891/api/dcim/devices/14/?format=json",
            "display": "dmi01-akron-sw01",
            "name": "dmi01-akron-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/sites/2/?format=json",
                "display": "DM-Akron",
                "name": "DM-Akron",
                "slug": "dm-akron",
            },
            "location": None,
            "rack": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/racks/1/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:11.625000Z",
        },
        {
            "id": 2,
            "url": "http://localhost:55891/api/dcim/devices/2/?format=json",
            "display": "dmi01-albany-rtr01",
            "name": "dmi01-albany-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/sites/3/?format=json",
                "display": "DM-Albany",
                "name": "DM-Albany",
                "slug": "dm-albany",
            },
            "location": None,
            "rack": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/racks/2/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 182,
                "url": "http://localhost:55891/api/ipam/ip-addresses/182/?format=json",
                "display": "2.2.2.2/10",
                "family": 4,
                "address": "2.2.2.2/10",
            },
            "primary_ip4": {
                "id": 182,
                "url": "http://localhost:55891/api/ipam/ip-addresses/182/?format=json",
                "display": "2.2.2.2/10",
                "family": 4,
                "address": "2.2.2.2/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:26:58.061000Z",
        },
        {
            "id": 15,
            "url": "http://localhost:55891/api/dcim/devices/15/?format=json",
            "display": "dmi01-albany-sw01",
            "name": "dmi01-albany-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/sites/3/?format=json",
                "display": "DM-Albany",
                "name": "DM-Albany",
                "slug": "dm-albany",
            },
            "location": None,
            "rack": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/racks/2/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:11.714000Z",
        },
        {
            "id": 3,
            "url": "http://localhost:55891/api/dcim/devices/3/?format=json",
            "display": "dmi01-binghamton-rtr01",
            "name": "dmi01-binghamton-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/sites/4/?format=json",
                "display": "DM-Binghamton",
                "name": "DM-Binghamton",
                "slug": "dm-binghamton",
            },
            "location": None,
            "rack": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/racks/3/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 183,
                "url": "http://localhost:55891/api/ipam/ip-addresses/183/?format=json",
                "display": "3.3.3.3/10",
                "family": 4,
                "address": "3.3.3.3/10",
            },
            "primary_ip4": {
                "id": 183,
                "url": "http://localhost:55891/api/ipam/ip-addresses/183/?format=json",
                "display": "3.3.3.3/10",
                "family": 4,
                "address": "3.3.3.3/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:27:13.871000Z",
        },
        {
            "id": 16,
            "url": "http://localhost:55891/api/dcim/devices/16/?format=json",
            "display": "dmi01-binghamton-sw01",
            "name": "dmi01-binghamton-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/sites/4/?format=json",
                "display": "DM-Binghamton",
                "name": "DM-Binghamton",
                "slug": "dm-binghamton",
            },
            "location": None,
            "rack": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/racks/3/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:11.799000Z",
        },
        {
            "id": 4,
            "url": "http://localhost:55891/api/dcim/devices/4/?format=json",
            "display": "dmi01-buffalo-rtr01",
            "name": "dmi01-buffalo-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/sites/5/?format=json",
                "display": "DM-Buffalo",
                "name": "DM-Buffalo",
                "slug": "dm-buffalo",
            },
            "location": None,
            "rack": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/racks/4/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 184,
                "url": "http://localhost:55891/api/ipam/ip-addresses/184/?format=json",
                "display": "4.4.4.4/10",
                "family": 4,
                "address": "4.4.4.4/10",
            },
            "primary_ip4": {
                "id": 184,
                "url": "http://localhost:55891/api/ipam/ip-addresses/184/?format=json",
                "display": "4.4.4.4/10",
                "family": 4,
                "address": "4.4.4.4/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:27:25.345000Z",
        },
        {
            "id": 17,
            "url": "http://localhost:55891/api/dcim/devices/17/?format=json",
            "display": "dmi01-buffalo-sw01",
            "name": "dmi01-buffalo-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/sites/5/?format=json",
                "display": "DM-Buffalo",
                "name": "DM-Buffalo",
                "slug": "dm-buffalo",
            },
            "location": None,
            "rack": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/racks/4/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:11.883000Z",
        },
        {
            "id": 5,
            "url": "http://localhost:55891/api/dcim/devices/5/?format=json",
            "display": "dmi01-camden-rtr01",
            "name": "dmi01-camden-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/sites/6/?format=json",
                "display": "DM-Camden",
                "name": "DM-Camden",
                "slug": "dm-camden",
            },
            "location": None,
            "rack": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/racks/5/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 185,
                "url": "http://localhost:55891/api/ipam/ip-addresses/185/?format=json",
                "display": "5.5.5.5/10",
                "family": 4,
                "address": "5.5.5.5/10",
            },
            "primary_ip4": {
                "id": 185,
                "url": "http://localhost:55891/api/ipam/ip-addresses/185/?format=json",
                "display": "5.5.5.5/10",
                "family": 4,
                "address": "5.5.5.5/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:27:37.376000Z",
        },
        {
            "id": 18,
            "url": "http://localhost:55891/api/dcim/devices/18/?format=json",
            "display": "dmi01-camden-sw01",
            "name": "dmi01-camden-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/sites/6/?format=json",
                "display": "DM-Camden",
                "name": "DM-Camden",
                "slug": "dm-camden",
            },
            "location": None,
            "rack": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/racks/5/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:11.968000Z",
        },
        {
            "id": 6,
            "url": "http://localhost:55891/api/dcim/devices/6/?format=json",
            "display": "dmi01-nashua-rtr01",
            "name": "dmi01-nashua-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/sites/7/?format=json",
                "display": "DM-Nashua",
                "name": "DM-Nashua",
                "slug": "dm-nashua",
            },
            "location": None,
            "rack": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/racks/6/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 186,
                "url": "http://localhost:55891/api/ipam/ip-addresses/186/?format=json",
                "display": "6.6.6.6/10",
                "family": 4,
                "address": "6.6.6.6/10",
            },
            "primary_ip4": {
                "id": 186,
                "url": "http://localhost:55891/api/ipam/ip-addresses/186/?format=json",
                "display": "6.6.6.6/10",
                "family": 4,
                "address": "6.6.6.6/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:27:46.894000Z",
        },
        {
            "id": 19,
            "url": "http://localhost:55891/api/dcim/devices/19/?format=json",
            "display": "dmi01-nashua-sw01",
            "name": "dmi01-nashua-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/sites/7/?format=json",
                "display": "DM-Nashua",
                "name": "DM-Nashua",
                "slug": "dm-nashua",
            },
            "location": None,
            "rack": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/racks/6/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.071000Z",
        },
        {
            "id": 7,
            "url": "http://localhost:55891/api/dcim/devices/7/?format=json",
            "display": "dmi01-pittsfield-rtr01",
            "name": "dmi01-pittsfield-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 8,
                "url": "http://localhost:55891/api/dcim/sites/8/?format=json",
                "display": "DM-Pittsfield",
                "name": "DM-Pittsfield",
                "slug": "dm-pittsfield",
            },
            "location": None,
            "rack": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/racks/7/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 187,
                "url": "http://localhost:55891/api/ipam/ip-addresses/187/?format=json",
                "display": "7.7.7.7/10",
                "family": 4,
                "address": "7.7.7.7/10",
            },
            "primary_ip4": {
                "id": 187,
                "url": "http://localhost:55891/api/ipam/ip-addresses/187/?format=json",
                "display": "7.7.7.7/10",
                "family": 4,
                "address": "7.7.7.7/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:27:59.599000Z",
        },
        {
            "id": 20,
            "url": "http://localhost:55891/api/dcim/devices/20/?format=json",
            "display": "dmi01-pittsfield-sw01",
            "name": "dmi01-pittsfield-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 8,
                "url": "http://localhost:55891/api/dcim/sites/8/?format=json",
                "display": "DM-Pittsfield",
                "name": "DM-Pittsfield",
                "slug": "dm-pittsfield",
            },
            "location": None,
            "rack": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/racks/7/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.204000Z",
        },
        {
            "id": 8,
            "url": "http://localhost:55891/api/dcim/devices/8/?format=json",
            "display": "dmi01-rochester-rtr01",
            "name": "dmi01-rochester-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/sites/9/?format=json",
                "display": "DM-Rochester",
                "name": "DM-Rochester",
                "slug": "dm-rochester",
            },
            "location": None,
            "rack": {
                "id": 8,
                "url": "http://localhost:55891/api/dcim/racks/8/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 188,
                "url": "http://localhost:55891/api/ipam/ip-addresses/188/?format=json",
                "display": "8.8.8.8/32",
                "family": 4,
                "address": "8.8.8.8/32",
            },
            "primary_ip4": {
                "id": 188,
                "url": "http://localhost:55891/api/ipam/ip-addresses/188/?format=json",
                "display": "8.8.8.8/32",
                "family": 4,
                "address": "8.8.8.8/32",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:28:11.364000Z",
        },
        {
            "id": 21,
            "url": "http://localhost:55891/api/dcim/devices/21/?format=json",
            "display": "dmi01-rochster-sw01",
            "name": "dmi01-rochster-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/sites/9/?format=json",
                "display": "DM-Rochester",
                "name": "DM-Rochester",
                "slug": "dm-rochester",
            },
            "location": None,
            "rack": {
                "id": 8,
                "url": "http://localhost:55891/api/dcim/racks/8/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.287000Z",
        },
        {
            "id": 9,
            "url": "http://localhost:55891/api/dcim/devices/9/?format=json",
            "display": "dmi01-scranton-rtr01",
            "name": "dmi01-scranton-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 10,
                "url": "http://localhost:55891/api/dcim/sites/10/?format=json",
                "display": "DM-Scranton",
                "name": "DM-Scranton",
                "slug": "dm-scranton",
            },
            "location": None,
            "rack": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/racks/9/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 189,
                "url": "http://localhost:55891/api/ipam/ip-addresses/189/?format=json",
                "display": "9.9.9.9/32",
                "family": 4,
                "address": "9.9.9.9/32",
            },
            "primary_ip4": {
                "id": 189,
                "url": "http://localhost:55891/api/ipam/ip-addresses/189/?format=json",
                "display": "9.9.9.9/32",
                "family": 4,
                "address": "9.9.9.9/32",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:28:23.927000Z",
        },
        {
            "id": 22,
            "url": "http://localhost:55891/api/dcim/devices/22/?format=json",
            "display": "dmi01-scranton-sw01",
            "name": "dmi01-scranton-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 10,
                "url": "http://localhost:55891/api/dcim/sites/10/?format=json",
                "display": "DM-Scranton",
                "name": "DM-Scranton",
                "slug": "dm-scranton",
            },
            "location": None,
            "rack": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/racks/9/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.370000Z",
        },
        {
            "id": 10,
            "url": "http://localhost:55891/api/dcim/devices/10/?format=json",
            "display": "dmi01-stamford-rtr01",
            "name": "dmi01-stamford-rtr01",
            "device_type": {
                "id": 6,
                "url": "http://localhost:55891/api/dcim/device-types/6/?format=json",
                "display": "ISR 1111-8P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "ISR 1111-8P",
                "slug": "isr1111",
            },
            "device_role": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/device-roles/1/?format=json",
                "display": "Router",
                "name": "Router",
                "slug": "router",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/platforms/1/?format=json",
                "display": "ios",
                "name": "ios",
                "slug": "ios",
            },
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 11,
                "url": "http://localhost:55891/api/dcim/sites/11/?format=json",
                "display": "DM-Stamford",
                "name": "DM-Stamford",
                "slug": "dm-stamford",
            },
            "location": None,
            "rack": {
                "id": 10,
                "url": "http://localhost:55891/api/dcim/racks/10/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 4,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": {
                "id": 190,
                "url": "http://localhost:55891/api/ipam/ip-addresses/190/?format=json",
                "display": "10.10.10.10/10",
                "family": 4,
                "address": "10.10.10.10/10",
            },
            "primary_ip4": {
                "id": 190,
                "url": "http://localhost:55891/api/ipam/ip-addresses/190/?format=json",
                "display": "10.10.10.10/10",
                "family": 4,
                "address": "10.10.10.10/10",
            },
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-20T00:00:00Z",
            "last_updated": "2022-06-15T07:28:37.781000Z",
        },
        {
            "id": 23,
            "url": "http://localhost:55891/api/dcim/devices/23/?format=json",
            "display": "dmi01-stamford-sw01",
            "name": "dmi01-stamford-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 11,
                "url": "http://localhost:55891/api/dcim/sites/11/?format=json",
                "display": "DM-Stamford",
                "name": "DM-Stamford",
                "slug": "dm-stamford",
            },
            "location": None,
            "rack": {
                "id": 10,
                "url": "http://localhost:55891/api/dcim/racks/10/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.454000Z",
        },
        {
            "id": 24,
            "url": "http://localhost:55891/api/dcim/devices/24/?format=json",
            "display": "dmi01-syracuse-sw01",
            "name": "dmi01-syracuse-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 12,
                "url": "http://localhost:55891/api/dcim/sites/12/?format=json",
                "display": "DM-Syracuse",
                "name": "DM-Syracuse",
                "slug": "dm-syracuse",
            },
            "location": None,
            "rack": {
                "id": 11,
                "url": "http://localhost:55891/api/dcim/racks/11/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.540000Z",
        },
        {
            "id": 25,
            "url": "http://localhost:55891/api/dcim/devices/25/?format=json",
            "display": "dmi01-utica-sw01",
            "name": "dmi01-utica-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 13,
                "url": "http://localhost:55891/api/dcim/sites/13/?format=json",
                "display": "DM-Utica",
                "name": "DM-Utica",
                "slug": "dm-utica",
            },
            "location": None,
            "rack": {
                "id": 12,
                "url": "http://localhost:55891/api/dcim/racks/12/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.622000Z",
        },
        {
            "id": 26,
            "url": "http://localhost:55891/api/dcim/devices/26/?format=json",
            "display": "dmi01-yonkers-sw01",
            "name": "dmi01-yonkers-sw01",
            "device_type": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-types/7/?format=json",
                "display": "C9200-48P",
                "manufacturer": {
                    "id": 3,
                    "url": "http://localhost:55891/api/dcim/manufacturers/3/?format=json",
                    "display": "Cisco",
                    "name": "Cisco",
                    "slug": "cisco",
                },
                "model": "C9200-48P",
                "slug": "c9200-48p",
            },
            "device_role": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/device-roles/4/?format=json",
                "display": "Access Switch",
                "name": "Access Switch",
                "slug": "access-switch",
            },
            "tenant": {
                "id": 5,
                "url": "http://localhost:55891/api/tenancy/tenants/5/?format=json",
                "display": "Dunder-Mifflin, Inc.",
                "name": "Dunder-Mifflin, Inc.",
                "slug": "dunder-mifflin",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 14,
                "url": "http://localhost:55891/api/dcim/sites/14/?format=json",
                "display": "DM-Yonkers",
                "name": "DM-Yonkers",
                "slug": "dm-yonkers",
            },
            "location": None,
            "rack": {
                "id": 13,
                "url": "http://localhost:55891/api/dcim/racks/13/?format=json",
                "display": "Comms closet",
                "name": "Comms closet",
            },
            "position": 10,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2020-12-22T00:00:00Z",
            "last_updated": "2020-12-22T02:11:12.705000Z",
        },
        {
            "id": 96,
            "url": "http://localhost:55891/api/dcim/devices/96/?format=json",
            "display": "ncsu-coreswitch1",
            "name": "ncsu-coreswitch1",
            "device_type": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/device-types/3/?format=json",
                "display": "EX9214",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "EX9214",
                "slug": "ex9214",
            },
            "device_role": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/device-roles/2/?format=json",
                "display": "Core Switch",
                "name": "Core Switch",
                "slug": "core-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 16,
                "url": "http://localhost:55891/api/dcim/racks/16/?format=json",
                "display": "R103",
                "name": "R103",
            },
            "position": 6,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-04-02T00:00:00Z",
            "last_updated": "2021-04-02T17:31:41.398000Z",
        },
        {
            "id": 97,
            "url": "http://localhost:55891/api/dcim/devices/97/?format=json",
            "display": "ncsu-coreswitch2",
            "name": "ncsu-coreswitch2",
            "device_type": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/device-types/3/?format=json",
                "display": "EX9214",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "EX9214",
                "slug": "ex9214",
            },
            "device_role": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/device-roles/2/?format=json",
                "display": "Core Switch",
                "name": "Core Switch",
                "slug": "core-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 17,
                "url": "http://localhost:55891/api/dcim/racks/17/?format=json",
                "display": "R104",
                "name": "R104",
            },
            "position": 6,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-04-02T00:00:00Z",
            "last_updated": "2021-04-02T17:31:56.519000Z",
        },
        {
            "id": 94,
            "url": "http://localhost:55891/api/dcim/devices/94/?format=json",
            "display": "ncsu117-distswitch1",
            "name": "ncsu117-distswitch1",
            "device_type": {
                "id": 12,
                "url": "http://localhost:55891/api/dcim/device-types/12/?format=json",
                "display": "QFX5110-48S-4C",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5110-48S-4C",
                "slug": "qfx5110-48s-4c",
            },
            "device_role": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/device-roles/3/?format=json",
                "display": "Distribution Switch",
                "name": "Distribution Switch",
                "slug": "distribution-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 22,
                "url": "http://localhost:55891/api/dcim/sites/22/?format=json",
                "display": "D. S. Weaver Labs",
                "name": "D. S. Weaver Labs",
                "slug": "ncsu-117",
            },
            "location": None,
            "rack": {
                "id": 40,
                "url": "http://localhost:55891/api/dcim/racks/40/?format=json",
                "display": "IDF117",
                "name": "IDF117",
            },
            "position": 36,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-04-02T00:00:00Z",
            "last_updated": "2021-04-02T17:29:22.719000Z",
        },
        {
            "id": 95,
            "url": "http://localhost:55891/api/dcim/devices/95/?format=json",
            "display": "ncsu118-distswitch1",
            "name": "ncsu118-distswitch1",
            "device_type": {
                "id": 12,
                "url": "http://localhost:55891/api/dcim/device-types/12/?format=json",
                "display": "QFX5110-48S-4C",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5110-48S-4C",
                "slug": "qfx5110-48s-4c",
            },
            "device_role": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/device-roles/3/?format=json",
                "display": "Distribution Switch",
                "name": "Distribution Switch",
                "slug": "distribution-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 23,
                "url": "http://localhost:55891/api/dcim/sites/23/?format=json",
                "display": "Grinnells Lab",
                "name": "Grinnells Lab",
                "slug": "ncsu-118",
            },
            "location": None,
            "rack": {
                "id": 38,
                "url": "http://localhost:55891/api/dcim/racks/38/?format=json",
                "display": "IDF118",
                "name": "IDF118",
            },
            "position": 36,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-04-02T00:00:00Z",
            "last_updated": "2021-04-02T17:29:56.066000Z",
        },
        {
            "id": 93,
            "url": "http://localhost:55891/api/dcim/devices/93/?format=json",
            "display": "ncsu128-distswitch1",
            "name": "ncsu128-distswitch1",
            "device_type": {
                "id": 12,
                "url": "http://localhost:55891/api/dcim/device-types/12/?format=json",
                "display": "QFX5110-48S-4C",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5110-48S-4C",
                "slug": "qfx5110-48s-4c",
            },
            "device_role": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/device-roles/3/?format=json",
                "display": "Distribution Switch",
                "name": "Distribution Switch",
                "slug": "distribution-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 24,
                "url": "http://localhost:55891/api/dcim/sites/24/?format=json",
                "display": "Butler Communications",
                "name": "Butler Communications",
                "slug": "ncsu-128",
            },
            "location": None,
            "rack": {
                "id": 39,
                "url": "http://localhost:55891/api/dcim/racks/39/?format=json",
                "display": "IDF128",
                "name": "IDF128",
            },
            "position": 32,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-04-02T00:00:00Z",
            "last_updated": "2021-04-02T17:28:42.498000Z",
        },
        {
            "id": 98,
            "url": "http://localhost:55891/api/dcim/devices/98/?format=json",
            "display": "r105-torsw:1 (98)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 18,
                "url": "http://localhost:55891/api/dcim/racks/18/?format=json",
                "display": "R105",
                "name": "R105",
            },
            "position": 26,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/1/?format=json",
                "display": "r105-torsw",
                "name": "r105-torsw",
                "master": {
                    "id": 98,
                    "url": "http://localhost:55891/api/dcim/devices/98/?format=json",
                    "display": "r105-torsw:1 (98)",
                    "name": None,
                },
            },
            "vc_position": 1,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2022-06-15T07:23:33.841000Z",
        },
        {
            "id": 99,
            "url": "http://localhost:55891/api/dcim/devices/99/?format=json",
            "display": "r105-torsw:2 (99)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 18,
                "url": "http://localhost:55891/api/dcim/racks/18/?format=json",
                "display": "R105",
                "name": "R105",
            },
            "position": 25,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/1/?format=json",
                "display": "r105-torsw",
                "name": "r105-torsw",
                "master": {
                    "id": 98,
                    "url": "http://localhost:55891/api/dcim/devices/98/?format=json",
                    "display": "r105-torsw:1 (98)",
                    "name": None,
                },
            },
            "vc_position": 2,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2021-12-30T16:16:55.811000Z",
        },
        {
            "id": 100,
            "url": "http://localhost:55891/api/dcim/devices/100/?format=json",
            "display": "r106-torsw:1 (100)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 19,
                "url": "http://localhost:55891/api/dcim/racks/19/?format=json",
                "display": "R106",
                "name": "R106",
            },
            "position": 26,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/2/?format=json",
                "display": "r106-torsw",
                "name": "r106-torsw",
                "master": {
                    "id": 100,
                    "url": "http://localhost:55891/api/dcim/devices/100/?format=json",
                    "display": "r106-torsw:1 (100)",
                    "name": None,
                },
            },
            "vc_position": 1,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2022-06-15T07:23:33.855000Z",
        },
        {
            "id": 101,
            "url": "http://localhost:55891/api/dcim/devices/101/?format=json",
            "display": "r106-torsw:2 (101)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 19,
                "url": "http://localhost:55891/api/dcim/racks/19/?format=json",
                "display": "R106",
                "name": "R106",
            },
            "position": 25,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/2/?format=json",
                "display": "r106-torsw",
                "name": "r106-torsw",
                "master": {
                    "id": 100,
                    "url": "http://localhost:55891/api/dcim/devices/100/?format=json",
                    "display": "r106-torsw:1 (100)",
                    "name": None,
                },
            },
            "vc_position": 2,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2021-12-30T16:17:27.958000Z",
        },
        {
            "id": 102,
            "url": "http://localhost:55891/api/dcim/devices/102/?format=json",
            "display": "r107-torsw:1 (102)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 20,
                "url": "http://localhost:55891/api/dcim/racks/20/?format=json",
                "display": "R107",
                "name": "R107",
            },
            "position": 26,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/3/?format=json",
                "display": "r107-torsw",
                "name": "r107-torsw",
                "master": {
                    "id": 102,
                    "url": "http://localhost:55891/api/dcim/devices/102/?format=json",
                    "display": "r107-torsw:1 (102)",
                    "name": None,
                },
            },
            "vc_position": 1,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2022-06-15T07:23:33.866000Z",
        },
        {
            "id": 103,
            "url": "http://localhost:55891/api/dcim/devices/103/?format=json",
            "display": "r107-torsw:2 (103)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 20,
                "url": "http://localhost:55891/api/dcim/racks/20/?format=json",
                "display": "R107",
                "name": "R107",
            },
            "position": 25,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 3,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/3/?format=json",
                "display": "r107-torsw",
                "name": "r107-torsw",
                "master": {
                    "id": 102,
                    "url": "http://localhost:55891/api/dcim/devices/102/?format=json",
                    "display": "r107-torsw:1 (102)",
                    "name": None,
                },
            },
            "vc_position": 2,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2021-12-30T16:17:56.487000Z",
        },
        {
            "id": 104,
            "url": "http://localhost:55891/api/dcim/devices/104/?format=json",
            "display": "r108-torsw:1 (104)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/racks/21/?format=json",
                "display": "R108",
                "name": "R108",
            },
            "position": 26,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/4/?format=json",
                "display": "r108-torsw",
                "name": "r108-torsw",
                "master": {
                    "id": 104,
                    "url": "http://localhost:55891/api/dcim/devices/104/?format=json",
                    "display": "r108-torsw:1 (104)",
                    "name": None,
                },
            },
            "vc_position": 1,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2022-06-15T07:23:33.882000Z",
        },
        {
            "id": 105,
            "url": "http://localhost:55891/api/dcim/devices/105/?format=json",
            "display": "r108-torsw:2 (105)",
            "name": None,
            "device_type": {
                "id": 5,
                "url": "http://localhost:55891/api/dcim/device-types/5/?format=json",
                "display": "QFX5100-48T-6Q",
                "manufacturer": {
                    "id": 7,
                    "url": "http://localhost:55891/api/dcim/manufacturers/7/?format=json",
                    "display": "Juniper",
                    "name": "Juniper",
                    "slug": "juniper",
                },
                "model": "QFX5100-48T-6Q",
                "slug": "qfx5100-48t-6q",
            },
            "device_role": {
                "id": 9,
                "url": "http://localhost:55891/api/dcim/device-roles/9/?format=json",
                "display": "ToR Switch",
                "name": "ToR Switch",
                "slug": "tor-switch",
            },
            "tenant": {
                "id": 13,
                "url": "http://localhost:55891/api/tenancy/tenants/13/?format=json",
                "display": "NC State University",
                "name": "NC State University",
                "slug": "nc-state",
            },
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 1,
                "url": "http://localhost:55891/api/dcim/locations/1/?format=json",
                "display": "Row 1",
                "name": "Row 1",
                "slug": "row-1",
                "_depth": 0,
            },
            "rack": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/racks/21/?format=json",
                "display": "R108",
                "name": "R108",
            },
            "position": 25,
            "face": {"value": "rear", "label": "Rear"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": {
                "id": 4,
                "url": "http://localhost:55891/api/dcim/virtual-chassis/4/?format=json",
                "display": "r108-torsw",
                "name": "r108-torsw",
                "master": {
                    "id": 104,
                    "url": "http://localhost:55891/api/dcim/devices/104/?format=json",
                    "display": "r108-torsw:1 (104)",
                    "name": None,
                },
            },
            "vc_position": 2,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2021-12-30T00:00:00Z",
            "last_updated": "2021-12-30T16:18:24.555000Z",
        },
        {
            "id": 106,
            "url": "http://localhost:55891/api/dcim/devices/106/?format=json",
            "display": "Lenovo Flex system (106)",
            "name": None,
            "device_type": {
                "id": 14,
                "url": "http://localhost:55891/api/dcim/device-types/14/?format=json",
                "display": "Flex system",
                "manufacturer": {
                    "id": 14,
                    "url": "http://localhost:55891/api/dcim/manufacturers/14/?format=json",
                    "display": "Lenovo",
                    "name": "Lenovo",
                    "slug": "lenovo",
                },
                "model": "Flex system",
                "slug": "flex-system",
            },
            "device_role": {
                "id": 7,
                "url": "http://localhost:55891/api/dcim/device-roles/7/?format=json",
                "display": "Application Server",
                "name": "Application Server",
                "slug": "application-server",
            },
            "tenant": None,
            "platform": None,
            "serial": "",
            "asset_tag": None,
            "site": {
                "id": 21,
                "url": "http://localhost:55891/api/dcim/sites/21/?format=json",
                "display": "MDF",
                "name": "MDF",
                "slug": "ncsu-065",
            },
            "location": {
                "id": 2,
                "url": "http://localhost:55891/api/dcim/locations/2/?format=json",
                "display": "Row 2",
                "name": "Row 2",
                "slug": "row-2",
                "_depth": 0,
            },
            "rack": {
                "id": 22,
                "url": "http://localhost:55891/api/dcim/racks/22/?format=json",
                "display": "R201",
                "name": "R201",
            },
            "position": 9,
            "face": {"value": "front", "label": "Front"},
            "parent_device": None,
            "status": {"value": "active", "label": "Active"},
            "airflow": None,
            "primary_ip": None,
            "primary_ip4": None,
            "primary_ip6": None,
            "cluster": None,
            "virtual_chassis": None,
            "vc_position": None,
            "vc_priority": None,
            "comments": "",
            "local_context_data": None,
            "tags": [],
            "custom_fields": {},
            "config_context": {},
            "created": "2022-01-10T00:00:00Z",
            "last_updated": "2022-01-10T21:00:04.030000Z",
        },
    ],
}

NETBOX_VMS_API_CALL = "/api/virtualization/virtual-machines/?format=json"
NETBOX_VMS_API_RESPONSE = {"count": 0,
                           "next": None, "previous": None, "results": []}

NETBOX_GENERATED_TESTBED = {
    "testbed": {"credentials": {"default": {"username": "test", "password": "test"}}},
    "devices": {
        "dmi01-albany-rtr01": {
            "os": "ios",
            "alias": "dmi01-albany-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "2.2.2.2"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-binghamton-rtr01": {
            "os": "ios",
            "alias": "dmi01-binghamton-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "3.3.3.3"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-buffalo-rtr01": {
            "os": "ios",
            "alias": "dmi01-buffalo-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "4.4.4.4"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-camden-rtr01": {
            "os": "ios",
            "alias": "dmi01-camden-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "5.5.5.5"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-nashua-rtr01": {
            "os": "ios",
            "alias": "dmi01-nashua-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "6.6.6.6"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-pittsfield-rtr01": {
            "os": "ios",
            "alias": "dmi01-pittsfield-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "7.7.7.7"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-rochester-rtr01": {
            "os": "ios",
            "alias": "dmi01-rochester-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "8.8.8.8"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-scranton-rtr01": {
            "os": "ios",
            "alias": "dmi01-scranton-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "9.9.9.9"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
        "dmi01-stamford-rtr01": {
            "os": "ios",
            "alias": "dmi01-stamford-rtr01",
            "platform": "ios",
            "type": "ISR 1111-8P",
            "connections": {"cli": {"protocol": "ssh", "ip": "10.10.10.10"}},
            "credentials": {"default": {"username": "test", "password": "test"}},
        },
    },
    "topology": {},
}



import os
import yaml
import re
import logging
import sys
import argparse

from pyats.utils.secret_strings import SecretString
from pyats.topology.loader.base import BaseTestbedLoader

logger = logging.getLogger(__name__)


class TestbedCreator(BaseTestbedLoader):
    """ TestbedCreator class (BaseTestbedLoader)

    The base testbed creator class. All testbed creators must inherit this class
    in order to properly integrate with pyATS CLI.

    You must override the '_generate' method in the derived class to specify the
    appropriate behaviour for your source. To add arguments to your derived
    class, simply override '_init_arguments' to return a dictionary of required
    or optional arguments. See the function for more information.

    Examples:
        # Example demonstrating the creation of a MySQL loader
        class Mysql(TestbedCreator):
            def _init_arguments(self):
                return {
                    "required": ["sql_username", "sql_password"]
                    "optional": {
                        "sql_table": "devices"
                    }
                }

            def _generate(self):
                # <Parsing Logic and Code>
                return testbed_data

        # Instantiation and usage
        creator = Mysql(sql_username='root', sql_password='admin')
        creator.to_testbed_file('tesbed.yaml')
        creator.to_testbed_object()
    
    """

    def __init__(self, **kwargs):
        """ Instantiates the testbed creator with appropriate arguments.
        
        """
        self._result = {'success': {}, 'errored':{}, 'warning':{}}
        self._keys = ['hostname','ip','username', 'password', 'protocol', 'os']
        self._cli_list_arguments = []
        self._cli_replacements = {}

        arguments = self._init_arguments()
        kwargs.update(self._parse_cli())

        if "required" in arguments:
            for arg in arguments["required"]:
                if arg in kwargs:
                    self.__dict__.setdefault('_' + arg, kwargs[arg])
                else:
                    required = '\n'.join(arguments["required"])
                    raise Exception(
                        "This following arguments are required for this source:"
                        "\n" + required + "\n\nSource Help:\n" + self.__doc__
                    )

        if "optional" in arguments:
            for arg in arguments["optional"]:
                self.__dict__.setdefault('_' + arg, kwargs[arg] 
                            if arg in kwargs else arguments["optional"][arg])

    def _parse_cli(self):
        """ Parses arguments from CLI if any. Removes the first two dashes and
            converts any left over dashes to underscores.

        Returns:
            dict: The parsed arguments in dictionary format.

        """
        parser = argparse.ArgumentParser()
        parser.add_argument('args', nargs=argparse.REMAINDER)
        args = parser.parse_args(sys.argv)
        kwargs = {}

        if args.args and len(args.args) > 0:
            args.args = args.args[1:]
            i = 0

            while i < len(args.args):
                arg = args.args[i]
                i += 1

                # If argument name is in replacement dictionary, 
                # replace it with correspoding name and value
                if arg in self._cli_replacements:
                    name, value = self._cli_replacements[arg]
                    kwargs.setdefault(name, value)
                    continue

                # If argument expects a list, search and return list
                if arg in self._cli_list_arguments:
                    j = i
                    
                    # Collect parameters
                    while j < len(args.args) and not args.args[j].startswith('--') and \
                        not '-r' in args.args[j]: j += 1

                    kwargs.setdefault(arg.replace('--', '').replace('-', '_'), 
                                                                args.args[i:j])

                    # Incrememt index
                    i = j
                    continue

                parts = arg.split('=', 1)
                key = None
                value = None

                if len(parts) < 2:
                    value = None

                    if i < len(args.args) and not args.args[i].startswith('-'):
                        # Handle spaces
                        value = args.args[i]
                        i += 1
                    else: 
                        # Assume flag value if no assignment is provided
                        value = True
                    
                    key = parts[0]
                else:
                    key, value = parts
                
                # Convert key to variable name
                key = key.replace('--', '')
                key = key.replace('-', '_')
                kwargs.setdefault(key, value)

        return kwargs

    def _init_arguments(self):
        """ Defines argument that should be added to the class instance. Should 
            be overridden in derived classes if they require arguments. This 
            will ensure correct arguments are passed through either the derived 
            constructor or the CLI.

            Return type must be a dict of maximum two keys: "required" and 
            "optional". Required key must pair with an array of argument names 
            which will be added to the class instance with a underscore in 
            front. Optional key must pair with a dictionary of argument names 
            and their default value. The following demonstrates an example:

                return {
                    "required": ["file_name"],
                    "optional": {
                        "encode_password": True
                    }
                }
        
            This will add 'self._file_name' and 'self._encode_password' property
            which the derived class can access.

            Adding to the 'self._cli_list_arguments' will signify the parser 
            that the argument expects a list. For example, to have '--key' to be
            parsed as a list in '--key a b c', you will need:
        
                self._cli_list_arguments.append('--key')

            Adding to the 'self._cli_replacements' will replace any occurrences 
            of the key with the corresponding variable name and value. For 
            example, to map '-s' argument to 'silent=True' variable, 
            you will need:

                self._cli_replacements.setdefault('-s', ('silent', True))

        Returns:
            dict: The arguments for the creator.

        """
        return {}

    def _generate(self):
        """ Defines the generate method that the derived class must implement. 
        
        Returns:
            dict: The intermediate dictionary format of the testbed data.

        """
        raise NotImplementedError(
            "Derived class must implement `_generate` method."
        )

    def to_testbed_file(self, output_location):
        """ Retrieves the testbed information and saves it as a YAML file.
        
        Args:
            output_location ('str'): Path of where to save the testbed file.
        
        Returns:
            bool: Indication if the operation is successful or not.

        """
        if os.path.isdir(output_location):
            raise Exception('Output "{o}" is a directory'
                                                    .format(o=output_location))
        testbed = self._generate()
        encode_password = False
        
        if hasattr(self, '_encode_password'):
            encode_password = self._encode_password

        try:
            self._write_yaml(output_location, testbed, encode_password)
            
            return True
        except:
            self._result['errored']['>'] = 'An error has occurred.'
            return False

    def load(self):
        """ Overrides the base testbed loader class to return a testbed object.

        Returns:
            Testbed: The testbed object.

        """
        return self.to_testbed_object()

    def to_testbed_object(self):
        """ Retrieves the testbed information and returns it as a testbed object.
        
        Returns:
            Testbed: The testbed object.

        """
        data = self._generate()

        if data is None:
            return None

        return self._create_testbed(data)

    def _create_testbed(self, data):
        """ Helper for creating testbed object from intermediate testbed
            dictionaries.

        Args:
            data ('dict'): The testbed data.

        Returns:
            Testbed: The converted testbed object.
        
        """
        return BaseTestbedLoader.create_testbed({
            'testbed': {
                'name': 'testbed'
            },
            'devices': data.get('devices', {}),
            'topology': data.get('topology', {})
        })

    def _encode_all_password(self, devices):
        """ Encode the password of all the devices.
        
        Args:
            devices ('dict'): The intermediate testbed dictionary.

        """
        # ask password on connect if not provided, otherwise encode the password
        stack = [devices]
        while len(stack) > 0:
            current = stack.pop()
            for key, value in current.items():
                if key == "password" and value != '%ASK{}':
                    value = self._encode_secret(value)
                    current[key] = value
                elif isinstance(value, dict):
                    stack.append(value)

    def _encode_secret(self, plain_text):
        """ Performs password encoding.

        Args:
            plain_text ('str'): the plain text password.

        Returns:
            str: The encoded password.

        """
        encoded = SecretString.from_plaintext(plain_text)
        return '%ENC{' + encoded.data + '}'

    def _write_yaml(self, output, devices, encode_password, input_file=None):
        """ Write device data to yaml file.
        
        Args:
            output ('str'): The output file path.
            devices ('list'): Dictionary containing device data.
            encode_password ('bool'): Flag for encoding passwords or not.
            input_file ('str'): The input file name, if any.
        
        """
        # if empty dict, do nothing
        if not devices:
            return
        # Make sure output file can be created
        try:
            os.makedirs(os.path.dirname(output), exist_ok=True)
        except FileNotFoundError:
            # If a file which does not contains a directory name
            pass

        if encode_password:
            self._encode_all_password(devices)

        with open(output, 'w') as f:
            try:
                yaml.dump(devices, f, default_flow_style=False)
            except Exception as e:
                self._result['errored'][
                    input_file.lstrip('./')
                ] = 'has an error: {e}'.format(e=str(e))
                return
        if input_file:
            name = input_file.lstrip('./')
            self._result['success'].setdefault(name, "")
            self._result['success'][name] += '-> {f}\n'.format(f=output)
        else:
            self._result['success'][output] = ''

    def _construct_yaml(self, devices):
        """ Construct list of dicts containing device data into nested yaml 
            structure.

        Args:
            devices ('list'): List of dict containing device attributes.

        Returns:
             dict: Testbed dictionary that's ready to be dumped into yaml.
    
        """
        yaml_dict = {
            'devices': {}
        }
        seen_hostnames = set()
        for row in devices:
            try:
                name = row.pop('hostname')
            except KeyError:
                raise KeyError('Empty line found in given CSV/Excel file.')

            if name in seen_hostnames:
                raise Exception('Duplicate hostname "{n}" detected'
                                                                .format(n=name))
            else:
                seen_hostnames.add(name)

            try:
                # get port from ip
                ad_port = row['ip'].strip().rsplit(':', 1)
                address, port = ad_port[0], ad_port[1] if len(ad_port) > 1 else None
                os = row.pop('os')

                # build the connection dict
                if port:
                    connections = {
                        'cli': {
                            'ip': address,
                            'port': int(port),
                            'protocol': row.pop('protocol')
                        }
                    }
                else:
                    connections = {
                        'cli': {
                            'ip': row.pop('ip'),
                            'protocol': row.pop('protocol')}}
                
                if 'proxy' in row:
                    connections['cli'].update({'proxy': row.pop('proxy')})

                # build the credentials dict
                password = row.pop('password', '%ASK{}')
                if 'enable_password' in self._keys:
                    enable_password = row.pop('enable_password', '%ASK{}')
                else:
                    enable_password = row.pop('enable_password', password)
                credentials = {
                    'default': {
                        'username': row.pop('username'),
                        'password': password},
                    'enable': {
                        'password':  enable_password
                    }}

            except KeyError as e:
                raise KeyError('Missing required key {k} for device {d}'
                                                    .format(k=str(e), d=name))
            dev = yaml_dict['devices'].setdefault(name, {})
            dev['os'] = os
            dev['connections'] = connections
            dev['credentials'] = credentials
            type = row.get('type')
            dev['type'] = type if type else os
            if row and len(row) > 0:
                for key, value in row.items():
                    if 'custom:' in key:
                        dev.setdefault('custom', {}).setdefault(
                            key.replace('custom:', ''), value)
                    else:
                        dev.setdefault(key, value)

        return yaml_dict

    def print_result(self):
        """ Prints the result of testbed creating process.

        """
        # print testbeds create unsuccessfully
        if not self._result['success'] and not self._result['errored'] \
            and not self._result['warning']:
            logger.warning('No file found.')
            return

        if self._result['success']:
            if 'template' in self._result['success']:
                # print template create successfully
                logger.info('Template file generated: {file}'.format(file=self._result['success']['template']))
            else:
                # print testbeds create successfully
                logger.info('Testbed file generated: ')
                for k, v in self._result['success'].items():
                    logger.info('{k} {v}'.format(k=k,v=v))

        # print the ones that are errored
        if self._result['errored']:
            logger.info('')
            logger.error('Errors:')
            for k, v in self._result['errored'].items():
                logger.error('{k} {v}'.format(k=k,v=v))

        # print warnings
        if self._result['warning']:
            logger.info('')
            logger.warning('Warnings:')
            for k, v in self._result['warning'].items():
                logger.warning('{k} {v}'.format(k=k,v=v))



import os
import xlrd
import csv
import pathlib

from pyats.topology import loader
from .creator import TestbedCreator

class File(TestbedCreator):
    """ File class (TestbedCreator)

    Creator for the 'file' source. Takes in a CSV or Excel file and outputs the
    corresponding testbed object or file. Alternatively, it can take in a folder
    as path and converts all the CSV and Excel files inside.

    Args:
        path ('str'): The path of the input CSV/Excel file or a folder.
        recurse ('bool') default=False: If a folder is passed in, whether or not 
            traversal should include subdirectories.
        encode_password ('bool') default=False: Should generated testbed encode 
            its passwords.

    CLI Argument        |  Class Argument
    ---------------------------------------------
    --path=value        |  path=value
    --encode-password   |  encode_password=True
    -r                  |  recurse=True

    pyATS Examples:
        pyats create testbed file --path=test.csv --output=testbed.yaml
        pyats create testbed file --path=folder --output=testbeds -r

    Examples:
        # Create testbed from test.csv with encoded password
        creator = File(path="test.csv", encode_password=True)
        creator.to_testbed_file("testbed.yaml")
        creator.to_testbed_object()

    """

    def _init_arguments(self):
        """ Specifies the arguments for the creator.

        Returns:
            dict: Arguments for the creator.

        """
        self._cli_replacements.setdefault('-r', ('recurse', True))
        
        return {
            'required': ['path'],
            'optional': {
                'recurse': False,
                'encode_password': False
            }
        }

    def to_testbed_file(self, output_location):
        """ Saves the source data as a testbed file.

        Args:
            output_location ('str'): Where to save the file.
        
        Returns:
            bool: Indication that the operation is successful or not.
        
        """
        testbed = self._generate()

        if isinstance(testbed, list):
            for base, item in testbed:
                self._write_yaml(os.path.join(output_location, base), 
                            item, self._encode_password, input_file=self._path)
        else:
            self._write_yaml(output_location, testbed, self._encode_password,
                                                        input_file=self._path)

        return True

    def to_testbed_object(self):
        """ Creates testbed object from the source data.
        
        Returns:
            Testbed: The created testbed.
        
        """
        testbed = self._generate()
        
        if isinstance(testbed, list):
            return [self._create_testbed(data) for _, data in testbed]
        else:
            return self._create_testbed(testbed)

    def _generate(self):
        """ Core implementation of how the testbed data is created.

        Returns: 
            dict: The intermediate testbed dictionary.

        """
        if not os.path.exists(self._path):
            raise FileNotFoundError('File or directory does not exist: %s' 
                                                                % self._path)
        
        # if is a dir then walk through it
        if os.path.isdir(self._path):
            result = []

            for root, _, files in os.walk(self._path):
                for file in files:
                    input_file = os.path.join(root, file)
                    relative = os.path.relpath(input_file, self._path)
                    devices = self._read_device_data(input_file)

                    # The testbed filename should be same as the file
                    output = os.path.splitext(relative)[0] + '.yaml'

                    result.append((output, self._construct_yaml(devices)))
            
                # if recursive option is not set, then stop after first level
                if not self._recurse:
                    break
        else:
            devices = self._read_device_data(self._path)
            return self._construct_yaml(devices)
        
        return result

    def _read_device_data(self, file):
        """ Read device data based on file type.

        Args:
            file ('str'): Path of the file.
        
        Returns:
            list: List of dictionaries containing device data.

        """
        _, extension = os.path.splitext(file)
        
        # Check if file is csv or xls
        devices = {}

        if extension == '.csv':
            devices = self._read_csv(file)
        elif extension in {'.xls', '.xlsx'}:
            devices = self._read_excel(file)
        else:
            raise Exception("Given path is not a folder or a CSV/Excel file.")

        return devices

    def _read_csv(self, file_name):
        """ Reads CSV file containing device data.

        Args:
            file_name ('str'): Name of the CSV file.

        Returns:
            list: List of dictionaries containing the device attributes from 
                each row of the file.

        """
        row_lst = []
        with open(file_name, 'r') as f:
            reader = csv.reader(f)
            self._keys = next(reader)
            for row in reader:
                # Only take key which has value
                row_lst.append({k: v for k, v in dict(
                                            zip(self._keys, row)).items() if v})
        return row_lst

    def _read_excel(self, file_name):
        """ Read Excel file containing device data.

        Args:
            file_name ('str'): name of the excel file

        Returns:
            list: List of dictionaries containing device attributes from each
                row of the file.

        """
        row_lst = []
        ws = xlrd.open_workbook(file_name).sheet_by_index(0)
        self._keys = ws.row_values(0)
        for i in range(1, ws.nrows):
            # Only take key which has value
            row_lst.append({k: v for k, v in dict(
                            zip(self._keys, ws.row_values(i))).items() if v})
        return row_lst



# Python
import logging
from time import sleep, time

# pyAts
from pyats.async_ import pcall
from pyats.log.utils import banner
from pyats.easypy.plugins.bases import BasePlugin

# Logger
log = logging.getLogger('ats.easypy.%s' % __name__)


class TopologyUpPlugin(BasePlugin):
    '''
    Runs before job starts, to verify virtual topology is up and running
    before executing the test script.
    '''

    @classmethod
    def configure_parser(cls, parser, legacy_cli = True):
        grp = parser.add_argument_group('TopologyUpPlugin')

        if legacy_cli:
            all_devices_up = ['-check_all_devices_up']
            connection_check_timeout = ['-connection_check_timeout']
            connection_check_interval = ['-connection_check_interval']
        else:
            all_devices_up = ['--check-all-devices-up']
            connection_check_timeout = ['--connection-check-timeout']
            connection_check_interval = ['--connection-check-interval']

        # -check_all_devices_up
        # --check-all-devices-up
        grp.add_argument(*all_devices_up,
                         dest='all_devices_up',
                         action="store_true",
                         help='Enable/Disable checking for topology up pre job execution')

        # -connection_check_timeout
        # --connection-check-timeout
        grp.add_argument(*connection_check_timeout,
                         dest='connection_check_timeout',
                         action='store',
                         default=120,
                         help='Total time allowed for checking devices connectivity')

        # -connection_check_interval
        # --connection-check-interval
        grp.add_argument(*connection_check_interval,
                         dest='connection_check_interval',
                         action='store',
                         default=10,
                         help='Time to sleep between device connectivity checks')

        return grp


    def pre_job(self, task):
        '''Try to connect to all the topology devices in parallel and make sure they
           are up and running before executing the test script.
        '''

        # Check for the argument controlling the plugin run (Checking devices)
        check_devices_up = self.runtime.args.all_devices_up

        if not check_devices_up:
            log.info("Checking all devices are up and ready is disabled, '--check-all-devices-up' "
                     "must be set to True in case of pyats runs or '-check_all_devices_up' set to "
                     "True in case of legacy easypy runs")
            return
        else:
            log.info("TopologyUp Plugin is enabled, will start the plugin checking for all "
                "the devices' connectivity!")

        # Set the timers
        start_time = time()
        timeout = self.runtime.args.connection_check_timeout
        interval = self.runtime.args.connection_check_interval

        log.info("Connectivity check timeout is '{timeout}' and "
            "connectivity check interval is '{interval}'".format(timeout=timeout, interval=interval))

        # check devices and exclude IXIA device
        devices_list = []
        for device in self.runtime.testbed.find_devices():
            if 'ixia' not in str(device.__class__):
                devices_list.append({'device': device})
            else:
                log.info("Device {device} is not supported and connected check is skipped.".format(device=device.name))

        # Trying to connect to all devices in parallel
        pcall_output = pcall(device_connect,
            ckwargs = {'start_time': start_time, 'timeout': timeout, 'interval': interval},
            ikwargs = devices_list)

        # Create Summary
        log.info(banner("Devices' connection trials summary"))

        failed_list = []
        succeeded_list = []

        for res, dev, count in pcall_output:
            if res is False:
                msg = "Device '{device}' connectivity check failed after '{count}' trial(s)".format(
                    device=dev, count=count)
                failed_list.append(msg)
            else:
                msg = "Device '{device}' connectivity passed after '{count}' trial(s)".format(
                    device=dev, count=count)
                succeeded_list.append(msg)

        for mes in succeeded_list:
            log.info(mes)

        for fail_mes in failed_list:
            if fail_mes == failed_list[0]:
                log.info('')
            log.warning(fail_mes)

        if failed_list:
            # Terminate testscript
            log.info(banner("TopologyUp Plugin end!"))
            raise Exception ("Not all the testbed devices are up and ready")
        else:
            log.info("All devices are up and ready, Connected succesfully!")
            log.info(banner("TopologyUp Plugin end!"))

        return


def device_connect(device, start_time, timeout, interval):
    '''Try to connect to the device and if fails, sleep for interval seconds and retry
       till the timeout is reached

        Args:
            device ('obj'): device to use
            start_time ('int'): Current time to calculate the timeout, seconds
            timeout ('int'): Timeout value when reached exit even if failed, seconds
            interval ('int'): Sleep time between retries, seconds

        Returns:
            result(`bool`): Device is successfully connected
            device.name(`str`): Device's name'
            count(`int`): Device's connectivity trials count

        Raises:
            None

    '''

    count = 0

    while (time() - start_time) < float(timeout):

        time_difference = time() - start_time

        count = count+1

        try:
            # Connect to the device
            device.connect()

        except:
            # Not ready sleep and retry
            log.info("Connecting to device '{device}' failed. Sleeping for '{interval}' seconds "\
                "and retry, remaining time {remaining_time}".format(
                device=device, interval=str(interval), remaining_time=str(float(timeout)-float(time_difference))))

            # Sleep for `interval` seconds
            sleep(int(interval))

            continue

        else:
            log.info("Successfully connected to '{device}'".format(device=device))

            # Return the pcall call with True
            return (True, device.name, count)

    return (False, device.name, count)


# entrypoint
topology_up_plugin = {
    'plugins': {
        'TopologyUpPlugin': {
            'class': TopologyUpPlugin,
            'enabled': True,
            'kwargs': {},
            'module': 'pyats.contrib.plugins.topoup_plugin.topoup',
            'name': 'TopologyUpPlugin'
        }
    }
}



from unittest import TestCase, main, mock
from pyats.contrib.creators.interactive import Interactive
from pyats.topology import Testbed
from pyats.datastructures import Configuration
from pyats.utils import secret_strings

class TestInteractive(TestCase):

    maxDiff = None
    # set default pyats configuration
    secret_strings.cfg = Configuration()

    @mock.patch('builtins.input')
    @mock.patch('getpass.getpass')
    def test_interactive(self, getpass, input_function):
        input_value = [
            "y", "admin", "y", "super", "y", "hostname",
            "123.123.123.123", "ssh", "ios", "n"
        ]
        value = input_value.copy()
        expected = """devices:
  hostname:
    connections:
      cli:
        ip: 123.123.123.123
        protocol: ssh
    credentials:
      default:
        password: super
        username: admin
      enable:
        password: super
    os: ios
    type: ios
"""
        def mock_input(text):
            return value.pop(0)
        input_function.side_effect = mock_input
        getpass.return_value = "super"
        output_file = '/tmp/test.yaml'
        Interactive().to_testbed_file(output_file)
        with open(output_file) as file:
            self.assertEqual(file.read(), expected)
        value = input_value.copy()
        testbed = Interactive().to_testbed_object()
        self.assertTrue(isinstance(testbed, Testbed))
        self.assertIn('hostname', testbed.devices)
        self.assertEqual(testbed.devices['hostname'].os, 'ios')
        self.assertEqual(testbed.devices['hostname'].type, 'ios')
        self.assertIn('cli', testbed.devices['hostname'].connections)
        self.assertEqual('123.123.123.123', 
                        testbed.devices['hostname'].connections.cli.ip)
        self.assertEqual('ssh', 
                        testbed.devices['hostname'].connections.cli.protocol)
        self.assertIn('default', testbed.devices['hostname'].credentials)
        self.assertIn('enable', testbed.devices['hostname'].credentials)
        self.assertEqual('admin', 
                    testbed.devices['hostname'].credentials.default.username)

    @mock.patch('builtins.input')
    @mock.patch('getpass.getpass')
    def test_encode_password(self, getpass, input_function):
        input_value = [
            "y", "admin", "y", "super", "y", "hostname",
            "123.123.123.123", "ssh", "ios", "n"
        ]
        expected = """devices:
  hostname:
    connections:
      cli:
        ip: 123.123.123.123
        protocol: ssh
    credentials:
      default:
        password: '%ENC{w6PDrsORw5nDpQ==}'
        username: admin
      enable:
        password: '%ENC{w6PDrsORw5nDpQ==}'
    os: ios
    type: ios
"""
        def mock_input(text):
            return input_value.pop(0)
        input_function.side_effect = mock_input
        getpass.return_value = "super"
        output_file = '/tmp/test.yaml'
        Interactive(encode_password=True).to_testbed_file(output_file)
        with open(output_file) as file:
            self.assertEqual(file.read(), expected)

    @mock.patch('builtins.input')
    @mock.patch('getpass.getpass')
    def test_add_keys(self, getpass, input_function):
        input_value = [
            "y", "admin", "y", "super", "y", "hostname",
            "123.123.123.123", "ssh", "ios", "opt1", "opt2", "opt3", "n"
        ]
        value = input_value.copy()
        expected = """devices:
  hostname:
    a: opt1
    b: opt2
    c: opt3
    connections:
      cli:
        ip: 123.123.123.123
        protocol: ssh
    credentials:
      default:
        password: super
        username: admin
      enable:
        password: super
    os: ios
    type: ios
"""
        def mock_input(text):
            return value.pop(0)
        input_function.side_effect = mock_input
        getpass.return_value = "super"
        output_file = '/tmp/test.yaml'
        Interactive(add_keys=['a', 'b', 'c']).to_testbed_file(output_file)
        with open(output_file) as file:
            self.assertEqual(file.read(), expected)
        expected = """devices:
  hostname:
    connections:
      cli:
        ip: 123.123.123.123
        protocol: ssh
    credentials:
      default:
        password: super
        username: admin
      enable:
        password: super
    custom:
      w: opt1
      ww: opt2
      www: opt3
    os: ios
    type: ios
"""
        value = input_value.copy()
        Interactive(add_custom_keys=['w', 'ww', 'www']).to_testbed_file(
                                                                    output_file)
        with open(output_file) as file:
            self.assertEqual(file.read(), expected)

    @mock.patch('builtins.input')
    @mock.patch('getpass.getpass')
    def test_two_devices(self, getpass, input_function):
        input_value = [
            "n", "n", "n", "dev1", "123.123.123.123", "superuser", 
            "telnet", "linux", "y", "dev2", "123.123.123.123", "superuser", 
            "telnet", "linux", "n"
        ]
        expected = """devices:
  dev1:
    connections:
      cli:
        ip: 123.123.123.123
        protocol: telnet
    credentials:
      default:
        password: super
        username: superuser
      enable:
        password: super
    os: linux
    type: linux
  dev2:
    connections:
      cli:
        ip: 123.123.123.123
        protocol: telnet
    credentials:
      default:
        password: super
        username: superuser
      enable:
        password: super
    os: linux
    type: linux
"""
        def mock_input(text):
            return input_value.pop(0)
        input_function.side_effect = mock_input
        getpass.return_value = "super"
        output_file = '/tmp/.yaml'
        Interactive().to_testbed_file(output_file)
        with open(output_file) as file:
            self.assertEqual(file.read(), expected)

if __name__ == '__main__':
    main()



try:
    from ansible.parsing.dataloader import DataLoader
    from ansible.inventory.manager import InventoryManager
    from ansible.cli.inventory import InventoryCLI
    from ansible import context
except Exception:
    raise ImportError("'ansible' package is not installed. Please install by running: pip install ansible")
from .creator import TestbedCreator

class Ansible(TestbedCreator):
    """ Ansible class (TestbedCreator)

    Creator for the 'ansible' source. Reads the given inventory in Ansible and 
    converts the data to a structured testbed file or object.

    Args:
        inventory_name ('str'): The name of the Ansible inventory.
        encode_password ('bool') default=False: Should generated testbed encode 
            its passwords.

    CLI Argument           |  Class Argument
    ------------------------------------------------
    --inventory-name=value |  inventory_name=value
    --encode-password      |  encode_password=True

    pyATS Examples:
        pyats create testbed ansible --output=out --inventory-name=inventory.ini

    Examples:
        # Create testbed from Ansible source
        creator = Ansible(inventory_name="inventory.ini")
        creator.to_testbed_file("template.csv")
        creator.to_testbed_object()

    """
    
    def _init_arguments(self):
        """ Specifies the arguments for the creator.

        Returns:
            dict: Arguments for the creator.

        """
        return {
            'required': ['inventory_name'],
            'optional': {
                'encode_password': False
            }
        }

    def _generate(self):
        """ Transforms Ansible data into testbed format.
        
        Returns:
            dict: The intermediate dictionary format of the testbed data.
    
        """
        # Set Ansible arguments for export
        context.CLIARGS = {}
        context.CLIARGS['export'] = True
        context.CLIARGS['basedir'] = '.'

        # Instantiate Ansible control objects
        inventory = InventoryManager(loader=DataLoader(), 
                                                sources=self._inventory_name)
        test = InventoryCLI(args=[''])
        group = inventory.groups.get('all')
        test.inventory = inventory

        # Fetch all the data associated with particular inventory
        result = test.json_inventory(top=group)
        testbed = {}
        devices = testbed.setdefault('devices', {})
        host_vars = result['_meta']['hostvars']

        # Remove items that are not device type in result
        del result['all']
        del result['_meta']

        for device_type, category in result.items():
            # If category does not contain vars or hosts, we skip
            if not 'vars' in category or not 'hosts' in category:
                continue

            for host in category['hosts']:
                cli_name = 'cli'

                # If netconf is defined as connection type, use that instead
                # of default CLI type
                if 'ansible_connection' in category['vars']:
                    if 'netconf' in category['vars']['ansible_connection']:
                        cli_name = 'netconf'

                # Construct connection fields and credentials
                device = devices.setdefault(host, {})
                connections = device.setdefault('connections', {
                     cli_name: {'protocol': 'ssh'}
                })
                cli = connections[cli_name]
                default = device.setdefault('credentials', {'default': {}})
                default = default['default']

                # set connection ip
                if host in host_vars and 'ansible_host' in host_vars[host]:
                    cli.setdefault('ip', host_vars[host]['ansible_host'])
                else:
                    cli.setdefault('ip', host)

                # set connection port
                if 'ansible_ssh_port' in category['vars']:
                    cli.setdefault('port', category['vars']['ansible_ssh_port'])

                password = None
                # Select the correct field name based on what is given
                if 'ansible_ssh_pass' in category['vars']:
                    password = 'ansible_ssh_pass'
                elif 'ansible_password' in category['vars']:
                    password = 'ansible_password'
                else:
                    # If password does not exist, skip over device
                    del devices[host]
                    continue
                
                # Set password and username
                default.setdefault('password', category['vars'][password])
                default.setdefault('username', category['vars']['ansible_user'])

                # If device has any other connection types, we also
                # set those respectively with their password
                if 'ansible_become_method' in category['vars'] and \
                    'ansible_become_pass' in category['vars']:
                    inner = connections.setdefault(
                        category['vars']['ansible_become_method'], {})
                    inner.setdefault('password',
                        category['vars']['ansible_become_pass'])

                # Set other device properties
                device.setdefault('alias', host)

                if 'ansible_network_os' not in category['vars']:
                    raise Exception("Missing key word 'ansible_network_os' for %s" % host)

                device.setdefault('os', category['vars']['ansible_network_os'])
                device.setdefault('platform',
                                        category['vars']['ansible_network_os'])
                device.setdefault('type', device_type)

        return testbed if len(testbed['devices']) > 0 else None



import json
import logging
import os
import socket

import requests

from pyats.configuration import configuration as cfg
from pyats.easypy.plugins.bases import BasePlugin

logger = logging.getLogger(__name__)

MESSAGE_URL = 'https://webexapis.com/v1/messages'

MESSAGE_TEMPLATE = """
## JOB RESULT REPORT
```
Job ID        : {job.uid}
Host          : {job.runtime.env[host][name]}
Archive       : {job.runtime.runinfo.archive_file}
Total Tasks   : {job.runtime.tasks.count}
Total Runtime : {job.elapsedtime}s

Results Summary
---------------
Passed        : {job.results[passed]}
Passx         : {job.results[passx]}
Failed        : {job.results[failed]}
Aborted       : {job.results[aborted]}
Blocked       : {job.results[blocked]}
Skipped       : {job.results[skipped]}
Errored       : {job.results[errored]}
```
"""

class WebExTeamsNotifyPlugin(BasePlugin):
    '''
    Runs after job, sends notification with results to a specified WebEx Teams
    space, or a specific person.
    '''

    @classmethod
    def configure_parser(cls, parser, legacy_cli = True):
        grp = parser.add_argument_group('WebEx')

        if legacy_cli:
            space = ['-webex_space']
            email = ['-webex_email']
            token = ['-webex_token']
        else:
            space = ['--webex-space']
            email = ['--webex-email']
            token = ['--webex-token']

        grp.add_argument(*token,
                         dest='webex_token',
                         action="store",
                         type=str,
                         metavar='',
                         default = None,
                         help='Webex Bot AUTH Token')

        grp.add_argument(*space,
                         dest='webex_space',
                         action="store",
                         type=str,
                         metavar='',
                         default = None,
                         help='Webex Space ID to send notification to')

        grp.add_argument(*email,
                         dest='webex_email',
                         action="store",
                         type=str,
                         metavar='',
                         default = None,
                         help='Email of specific user to send WebEx '
                              'notification to')
        return grp


    def post_job(self, job):
        # Get WebEx info from arguments or configuration
        token = self.runtime.args.webex_token or cfg.get('webex.token')
        space = self.runtime.args.webex_space or cfg.get('webex.space')
        email = self.runtime.args.webex_email or cfg.get('webex.email')

        if not token:
            logger.info('WebEx Token not given as argument or in config. No '
                        'WebEx notification will be sent')
            return

        headers = {'Authorization': 'Bearer {}'.format(token),
                   'Content-Type': 'application/json'}

        if not space and not email:
            logger.info('No Space ID or email specified, No WebEx Teams '
                        'notification will be sent')
            return

        # Format message with info from job run
        msg = MESSAGE_TEMPLATE.format(job=job)
        
        # internal Cisco pyATS log upload link
        # (does not exist for external release)
        try:
            # Attempt to get path for TRADe logs
            if not self.runtime.runinfo.no_upload:
                msg += '\n\nView pyATS logs at: %s'\
                       % self.runtime.runinfo.log_url
        except AttributeError:
            pass

        try:
            host = job.runtime.env['host']['name']
            # Determine if liveview is running
            if self.runtime.args.liveview and\
                    self.runtime.args.liveview_keepalive:
                # Liveview will set this to the assigned port if not specified
                port = self.runtime.args.liveview_port
                try:
                    # Attempt to add a link using the host domain name
                    addr = socket.getfqdn()
                    socket.gethostbyname(addr)
                    msg += '\n\nLogs can be viewed with the pyATS Log Viewer '\
                           'at: http://%s:%s' % (addr, port)
                except OSError:
                    msg += '\n\nLogs can be viewed in your browser by '\
                           'connecting to %s with port %s' % (host, port)
            else:
                # Show command to run liveview
                archive = self.runtime.runinfo.archive_file
                if archive:
                    msg += '\n\nRun the following command on %s to view logs '\
                           'from this job: `pyats logs view %s --host 0.0.0.0`'\
                           % (host, archive)

        except AttributeError:
            pass

        # Build payload
        payload = {'markdown': msg}
        if space:
            payload['roomId'] = space
        elif email:
            payload['toPersonEmail'] = email

        logger.info('Sending WebEx Teams notification')

        try:
            # Attempt POST
            r = requests.post(MESSAGE_URL,
                              data=json.dumps(payload),
                              headers=headers)
            logger.debug('notification status: %s' % r.status_code)
            logger.debug(r.text)
        except Exception:
            logger.exception('Failed to send WebEx Teams notification:')


# entrypoint
webex_plugin = {
    'plugins': {
        'WebExTeamsNotifyPlugin': {
            'class': WebExTeamsNotifyPlugin,
            'enabled': True,
            'kwargs': {},
            'module': 'pyats.contrib.plugins.webex_plugin.webex',
            'name': 'WebExTeamsNotifyPlugin'
        }
    }
}



import logging
import getpass
import os

from .creator import TestbedCreator

logger = logging.getLogger(__name__)

class Interactive(TestbedCreator):
    """ Interactive class (TestbedCreator)

    Creator for the 'interactive' source. Runs a console script to prompt step 
    by step questions and generating the corresponding testbed object or file 
    as an output.

    Args:
        add_keys ('list') default=None: Any additional keys that should be added
            to the testbed.
        encode_password ('bool') default=False: Should generated testbed encode 
            its passwords.

    CLI Argument                |  Class Argument
    -----------------------------------------------------------------
    --add-keys k1 k2 ...        |  add_keys=['k1', 'k2', ...]
    --add-custom-keys k1 k2 ... |  add_custom_keys=['k1', 'k2', ...]
    --encode-password           |  encode_password=True

    pyATS Examples:
        pyats create testbed interactive --output=testbed.yaml

    Examples:
        # Create testbed from CLI with additional keys
        creator = Interactive(encode_password=True, add_keys=['time', 'date'])
        creator.to_testbed_file("testbed.yaml")
        creator.to_testbed_object()

    """

    _VALID_ANSWER = {'yes': True, 'y': True, 'no': False, 'n': False}

    def _init_arguments(self):
        """ Specifies the arguments for the creator.

        Returns:
            dict: Arguments for the creator.

        """
        self._cli_list_arguments.append('--add-keys')
        self._cli_list_arguments.append('--add-custom-keys')

        return {
            'optional': {
                'encode_password': False,
                'add_keys': None,
                'add_custom_keys': None
            }
        }

    def _prompt_password(self, msg):
        """ Prompts user to enter password, and set to '%ASK{}' if nothing 
            entered.

        Returns:
            str: The password that the user entered.

        """
        password = getpass.getpass(prompt=msg)
        if not password:
            return '%ASK{}'
        else:
            return password

    def _get_info(self, msg, iterable=None, invalid=False):
        """ Prompts input from user, validating the input if iterable is
            provided.

        Args:
            msg ('str'): The prompt message.
            iterable ('iterable'): Iterable object that contains the 
                valid/invalid answers.
            invalid ('bool'): Flag that tells if iterable is invalid answer.

        Returns:
            str: The user's input.

        """
        response = ''

        # if provided a list of valid answers, then check if input is valid
        if iterable:
            if invalid:
                while response in iterable:
                    response = input(msg)
            else:
                while response not in iterable:
                    response = input(msg)
        else:
            response = input(msg)
        return response

    def _generate(self):
        """ Core implementation of how the testbed data is created.

        Returns: 
            dict: The intermediate testbed dictionary.

        """
        # Make sure output isn't a directory
        logger.info('Start creating Testbed yaml file ...')
        devices = []
        more_device = True
        all_password = None
        all_username = None
        enable_all_password = None
        name_set = set()

        # (keyname, description)
        keys = [('hostname', ''),
                ('IP', '(ip, or ip:port)'),
                ('Username', ''),
                ('Password', ''),
                ('Protocol', '(ssh, telnet, ...)'),
                ('OS', '(iosxr, iosxe, ios, nxos, linux, ...)')]

        # check if all devices have same username
        user_answer = self._get_info(
            'Do all of the devices have the same username? [y/n] ',
                                                    iterable=self._VALID_ANSWER)

        # if same username, ask user
        if self._VALID_ANSWER.get(user_answer):
            all_username = self._get_info(
                            'Common Username: ', iterable={''}, invalid=True)
            logger.info('')

        # check if all devices have same password
        pass_answer = self._get_info(
            'Do all of the devices have the same default password? [y/n] ',
                                                    iterable=self._VALID_ANSWER)

        # if same password, ask user
        if self._VALID_ANSWER.get(pass_answer):
            all_password = self._prompt_password(
                "Common Default Password "
                + "(leave blank if you want to enter on demand): ")
            logger.info('')

        enable_pass_answer = self._get_info(
            'Do all of the devices have the same enable password? [y/n] ',
            iterable=self._VALID_ANSWER)

        # if same enable password, ask, if nothing is entered, 
        # set it to the same as default password
        if self._VALID_ANSWER.get(enable_pass_answer):
            enable_all_password = self._prompt_password(
                "Common Enable Password " + 
                "(leave blank if you want to enter on demand): ")
            logger.info('')

        while more_device:
            logger.info('')
            device = {}
            for key, description in keys:

                # Get Device hostname
                if key == 'hostname':
                    name = self._get_info('Device hostname: ', iterable={''}, 
                                                                invalid=True)

                    # if device name already exist, ask again
                    while name in name_set:
                        logger.info('{d} has been already entered'
                                                                .format(d=name))
                        name = self._get_info('Device hostname: ', 
                                                    iterable={''}, invalid=True)
                    device['hostname'] = name
                    name_set.add(name)
                    continue

                elif key.lower() == 'username':

                    # ask user for username
                    if all_username:
                        device['username'] = all_username
                    else:
                        device['username'] = self._get_info(
                                '   Username: ', iterable={''}, invalid=True)
                    continue

                elif key.lower() == 'password':

                    # ask user for password if not all devices has the same
                    if all_password:
                        device['password'] = all_password
                    else: 
                        device['password'] = self._prompt_password(
                        "Default Password " +
                        "(leave blank if you want to enter on demand): ")
                    # ask for enable password if not the same, 
                    # and if user entered nothing set default password to enable
                    if enable_all_password:
                        enable_pass = enable_all_password 
                    else:
                        enable_pass = self._prompt_password(
                        "Enable Password " +
                        "(leave blank if you want to enter on demand): ")

                    if enable_pass:
                        device['enable_password'] = enable_pass
                    else:
                        device['enable_password'] = ''
                    continue

                else:
                    # Any other key
                    device[key.lower()] = self._get_info('   {k} {d}: '
                    .format(k=key, d=description), iterable={''}, invalid=True)

            # ask input for custom keys if supplied
            if self._add_custom_keys:
                if not self._add_keys:
                    self._add_keys = []
                self._add_keys.extend("custom:{}".format(key.lower())
                                            for key in self._add_custom_keys)

            if self._add_keys:
                for k in self._add_keys:
                    if k not in device:
                        device[k.lower()] = self._get_info(
                            '   Value for custom key "{k}": '.format(k=k),
                            iterable={''},
                            invalid=True)

            devices.append(device)

            # ask if user want to enter more devices
            answer = self._get_info('More devices to add ? [y/n] ', 
                                            iterable=self._VALID_ANSWER).lower()
            more_device = self._VALID_ANSWER[answer]

        return self._construct_yaml(devices)




import os
import sys

from pyats.contrib.creators.creator import TestbedCreator
from unittest import TestCase, main
from pyats.topology import Testbed
from pyats.topology.loader.base import BaseTestbedLoader

class TestCreator(TestCase):
    def test_arguments(self):
        class Test(TestbedCreator):
            def _init_arguments(self):
                return {
                    "required": ["a"],
                    "optional": {
                        "b" : 12,
                        "c": [1, 2, 4]
                    }
                }
        with self.assertRaises(Exception):
            Test()
        with self.assertRaises(Exception):
            Test(b=2, c=4)
        test = Test(a=5)
        self.assertTrue(hasattr(test, '_a'))
        self.assertTrue(hasattr(test, '_b'))
        self.assertTrue(hasattr(test, '_c'))
        self.assertEqual(test._a, 5)
        self.assertEqual(test._b, 12)
        self.assertEqual(test._c, [1, 2, 4])
        test = Test(a=1, b=2, c=3)
        self.assertEqual(test._a, 1)
        self.assertEqual(test._b, 2)
        self.assertEqual(test._c, 3)

    def test_generate(self):
        with self.assertRaises(NotImplementedError):
            TestbedCreator()._generate()
        class Test(TestbedCreator):
            pass
        with self.assertRaises(NotImplementedError):
            Test().to_testbed_file('/tmp/test')
        with self.assertRaises(NotImplementedError):
            Test().to_testbed_object()
        
    def test_output_to_folder(self):
        folder = '/tmp/testfolder'
        if not os.path.isdir(folder):
            os.mkdir(folder)
        with self.assertRaises(Exception):
            TestbedCreator().to_testbed_file(folder)

    def test_cli_arguments(self):
        class Test(TestbedCreator):
            def _init_arguments(self):
                return {
                    "required": ["my_arg", "password"],
                    "optional": {
                        "b" : 'bc',
                        "time": 'on'
                    }
                }
        sys.argv = ["creator"]
        with self.assertRaises(Exception):
            Test()
        sys.argv = ["creator", "--my-arg=abc"]
        with self.assertRaises(Exception):
            Test()
        sys.argv = ["creator", "--my-arg=abc", "--password=admin"]
        test = Test()
        self.assertEqual(test._my_arg, 'abc')
        self.assertEqual(test._password, 'admin')
        self.assertEqual(test._b, 'bc')
        self.assertEqual(test._time, 'on')
        sys.argv = ["creator", "--my-arg", "--password"]
        test = Test()
        self.assertTrue(test._my_arg)
        self.assertTrue(test._password)
        sys.argv = ["creator", "--my-arg", "123", "--password", "abc"]
        test = Test()
        self.assertEqual(test._my_arg, '123')
        self.assertEqual(test._password, 'abc')
        sys.argv = ["creator", "--my-arg=a", "--password", "b"]
        test = Test()
        self.assertEqual(test._my_arg, 'a')
        self.assertTrue(test._password, 'b')
        sys.argv = ["creator", "--my-arg=a", "--password=b",
                                                        "--b=hi", "--time=55"]
        test = Test()
        self.assertEqual(test._my_arg, 'a')
        self.assertEqual(test._password, 'b')
        self.assertEqual(test._b, 'hi')
        self.assertEqual(test._time, '55')

    def test_mixed_arguments(self):
        class Test(TestbedCreator):
            def _init_arguments(self):
                return {
                    "required": ["my_arg", "password"]
                }
        sys.argv = ["creator", "--my-arg=abc"]
        test = Test(password="123")
        self.assertEqual(test._password, "123")
        self.assertEqual(test._my_arg, "abc")

    def test_cli_list(self):
        class Test(TestbedCreator):
            def _init_arguments(self):
                self._cli_list_arguments.append('--items')
                return {
                    "required": ["items", "a", "b"]
                }
        sys.argv = ["creator", "--items", "1", "2", "3", "--a=1", "--b=1"]
        test = Test()
        self.assertEqual(test._items, ['1', '2', '3'])
        self.assertEqual(test._a, '1')
        self.assertEqual(test._b, '1')
        sys.argv = ["creator", "--a=1", "--items", "1", "2", "--b=1"]
        test = Test()
        self.assertEqual(test._items, ['1', '2'])
        self.assertEqual(test._a, '1')
        self.assertEqual(test._b, '1')
        sys.argv = ["creator", "--a=1", "--b=1", "--items", "aa", "bc", "cc"]
        test = Test()
        self.assertEqual(test._items, ['aa', 'bc', 'cc'])
        self.assertEqual(test._a, '1')
        self.assertEqual(test._b, '1')
        sys.argv = ["creator", "--a", "1", "--b", "1", "--items", "aa", "bc", "cc"]
        test = Test()
        self.assertEqual(test._items, ['aa', 'bc', 'cc'])
        self.assertEqual(test._a, '1')
        self.assertEqual(test._b, '1')
        sys.argv = ["creator", "--a=1", "--items", "--b=1"]
        test = Test()
        self.assertEqual(test._items, [])
        self.assertEqual(test._a, '1')
        self.assertEqual(test._b, '1')

    def test_cli_replacements(self):
        class Test(TestbedCreator):
            def _init_arguments(self):
                self._cli_replacements.setdefault('-r', ('recurse', True))
                self._cli_replacements.setdefault('--hello', ('h', 'w'))
                return {
                    "required": ["recurse"],
                    "optional": {
                        "h": "abc"
                    }
                }
        sys.argv = ["creator", "-r"]
        test = Test()
        self.assertTrue(test._recurse)
        self.assertEqual(test._h, "abc")
        sys.argv = ["creator", "-r", "--hello"]
        test = Test()
        self.assertEqual(test._h, 'w')

    def test_inheritance(self):
        self.assertTrue(issubclass(TestbedCreator, BaseTestbedLoader))

    def test_to_testbed_file_none(self):
        output = '/tmp/test'
        class Test(TestbedCreator):
            def _generate(self):
                return None
        if os.path.isfile(output):
            os.remove(output)
        Test().to_testbed_file(output)
        self.assertFalse(os.path.isfile(output))

    def test_to_testbed_object(self):
        class Test(TestbedCreator):
            def _generate(self):
                return None
        self.assertIsNone(Test().to_testbed_object())
        class Test(TestbedCreator):
            def _generate(self):
                return {}
        self.assertTrue(isinstance(Test().to_testbed_object(), Testbed))

    def test_load(self):
        class Test(TestbedCreator):
            def _generate(self):
                return {}
        self.assertTrue(isinstance(Test().to_testbed_object(), Testbed))

if __name__ == '__main__':
    main()        
